import os
import random
import string
from warnings import catch_warnings, resetwarnings

import pytest
from lya import AttrDict

from yaycl import Config, ConfigTree, ConfigNotFound, ConfigInvalid

test_yaml_contents = '''
test_key: test_value
nested_test_root:
    nested_test_key_1: nested_test_value_1
    nested_test_key_2: nested_test_value_2
inherit_value:
    inherit: nested_test_root
    inherit_nested_value:
        inherit: inherit_value
'''

local_test_yaml_contents = '''
test_key: test_overridden_value
nested_test_root:
    nested_test_key_1: overridden_nested_test_value
'''

invalid_test_yaml_contents = '''
% this isn't valid yaml
'''


@pytest.fixture
def clear_conf(conf):
    # Ensure the conf is cleared before every test
    conf.clear()
    conf.runtime.clear()
pytestmark = pytest.mark.usefixtures('clear_conf')


@pytest.fixture
def random_string():
    return ''.join((random.choice(string.letters) for _ in xrange(8)))


@pytest.fixture
def test_yaml(request, random_string):
    test_yaml = create_test_yaml(request, conf, test_yaml_contents, random_string)
    filename, ext = os.path.splitext(os.path.basename(test_yaml.name))
    return filename


@pytest.fixture(scope='session')
def conf_dir(request):
    dir = request.session.fspath.join('tests', 'conf')
    dir.ensure(dir=True)
    request.addfinalizer(lambda: delete_path(dir))
    return dir


@pytest.fixture
def conf(conf_dir):
    return Config(conf_dir.strpath)

def create_test_yaml(request, conf, contents, filename, local=False):
    if local:
        suffix = '.local.yaml'
    else:
        suffix = '.yaml'
    filename += suffix

    full_path = conf_dir(request).join(filename)

    test_yaml = full_path.open('w')
    test_yaml.write(contents)
    test_yaml.flush()
    test_yaml.seek(0)

    request.addfinalizer(lambda: delete_path(full_path))

    return test_yaml


def delete_path(path):
    if path.check():
        path.remove()

def test_conf_basics(test_yaml, conf):
    # Dict lookup method works
    assert conf[test_yaml]['test_key'] == 'test_value'
    assert isinstance(conf, Config)
    assert isinstance(conf.runtime, ConfigTree)
    assert isinstance(conf[test_yaml], AttrDict)
    assert isinstance(conf[test_yaml]['nested_test_root'], AttrDict)
    assert isinstance(conf[test_yaml].nested_test_root, AttrDict)
    assert conf[test_yaml]['nested_test_root'] is conf[test_yaml].nested_test_root


def test_conf_yamls_item(test_yaml, conf):
    # delitem doesn't really delete, it only clears in-place
    old_test_yaml = conf[test_yaml]
    del(conf[test_yaml])
    assert conf[test_yaml] is old_test_yaml

    # setitem doesn't really set, it only updates in-place
    conf[test_yaml] = {'foo': 'bar'}
    assert conf[test_yaml] is old_test_yaml


def test_conf_yamls_attr(test_yaml, conf):
    # Attribute lookup method works
    assert getattr(conf, test_yaml)['test_key'] == 'test_value'


def test_conf_yamls_save(request, test_yaml, conf):
    save_test = conf_dir(request).join('{}.yaml'.format('save_test'))
    request.addfinalizer(lambda: delete_path(save_test))
    with create_test_yaml(request, conf, local_test_yaml_contents, test_yaml):
        assert not save_test.check()
        conf['save_test'].update(conf[test_yaml])
        conf.save('save_test')
        del(conf['save_test'])
        assert save_test.check
        assert sorted(conf['save_test']) == sorted(conf[test_yaml])


def test_conf_yamls_override(request, test_yaml, conf):
    # Make a .local.yaml file with the same root name as test_yaml,
    with create_test_yaml(request, conf, local_test_yaml_contents, test_yaml, local=True):
        # Check that the simple local override works.
        assert conf[test_yaml]['test_key'] == 'test_overridden_value'

        # Check that the local override of specific nested keys works.
        nested_root = conf[test_yaml]['nested_test_root']
        assert nested_root['nested_test_key_1'] == 'overridden_nested_test_value'
        assert nested_root['nested_test_key_2'] == 'nested_test_value_2'


def test_conf_yamls_import(test_yaml, conf):
    # Emulate from conf import $test_yaml
    assert getattr(conf, test_yaml) == conf[test_yaml]


def test_conf_yamls_not_found(random_string, conf, recwarn):
    # Make sure the the ConfigNotFound warning is issued correctly
    resetwarnings()
    conf[random_string]
    assert recwarn.pop(ConfigNotFound)


def test_conf_doesnt_load_privates(conf):
    with pytest.raises(AttributeError):
        assert conf._private


def test_conf_runtime_override(random_string, conf):
    # Add the random string to the runtime dict, as well as a junk value to
    # prove we can add more than one thing via the ConfTree
    conf.runtime['test_config'][random_string] = True
    conf.runtime['foo'] = 'bar'
    # the override is in place
    assert random_string in conf.test_config
    conf.clear()
    # the override is still in place
    assert random_string in conf.test_config
    # both item and attr access works
    assert conf.test_config[random_string] is True
    assert getattr(conf.test_config, random_string) is True
    # deleting works
    del(conf.runtime['test_config'][random_string])
    assert random_string not in conf.test_config


def test_conf_runtime_set(random_string, conf):
    # setting runtime directly works, and doesn't change the object that runtime points to
    old_runtime = conf.runtime
    conf.runtime = {'test_config': {random_string: True}}
    conf.runtime is old_runtime
    assert random_string in conf.test_config


def test_conf_runtime_update(random_string, conf):
    # In addition to direct nested assignment, dict update should also work
    conf.runtime.update({'test_config': {random_string: True}})
    assert random_string in conf.test_config


def test_conf_runtime_clear(test_yaml, random_string, conf):
    conf.runtime[test_yaml]['test_key'] = random_string
    assert conf[test_yaml]['test_key'] == random_string
    conf.runtime.clear()
    assert conf[test_yaml]['test_key'] == 'test_value'


def test_conf_runtime_del(test_yaml, random_string, conf):
    conf.runtime[test_yaml]['test_key'] = random_string
    assert conf[test_yaml]['test_key'] == random_string
    del(conf.runtime)
    assert conf[test_yaml]['test_key'] == 'test_value'


def test_conf_imported_attr(test_yaml, random_string, conf):
    # "from conf import attr" should be the same object as "conf.attr"
    # mutating the runtime dict shouldn't change that
    imported_test_yaml = getattr(conf, test_yaml)
    conf.runtime[test_yaml]['test_key'] = random_string
    assert imported_test_yaml['test_key'] == random_string
    assert imported_test_yaml is getattr(conf, test_yaml)


def test_conf_override_before_import(test_yaml, random_string, conf):
    # You should be able to create a "fake" config file by preseeding the runtime overrides
    conf.runtime['foo']['test_key'] = random_string
    foo = getattr(conf, 'foo')
    assert foo['test_key'] == random_string


def test_conf_load_invalid_yaml(request, test_yaml, random_string, conf, recwarn):
    resetwarnings()
    with create_test_yaml(request, conf, invalid_test_yaml_contents, test_yaml, local=True):
        conf[test_yaml]
        assert recwarn.pop(ConfigInvalid)


def test_inheritance(request, test_yaml, conf):
    with create_test_yaml(request, conf, local_test_yaml_contents, test_yaml, local=True):
        # Check that nested iheritance (and thus inheritance itself) works
        assert (conf[test_yaml].nested_test_root ==
            conf[test_yaml].inherit_value.inherit_nested_value)


def test_impersonate_module(request, test_yaml, conf):
    import importlib
    import sys
    with create_test_yaml(request, conf, local_test_yaml_contents, test_yaml, local=True):
        # put our module-to-be in sys.modules
        sys.modules['conf'] = conf
        # simulate 'from conf import test_yaml'
        conf_module = importlib.import_module('conf')
        assert getattr(conf_module, test_yaml)



# -*- coding: utf-8 -*-
from collections import defaultdict
from functools import partial
from warnings import catch_warnings, warn

import os
import yaml
from pkg_resources import iter_entry_points

from lya import AttrDict, OrderedDictYAMLLoader


class ConfigNotFound(UserWarning):
    pass


class ConfigInvalid(UserWarning):
    pass


class Config(dict):
    """Configuration YAML loader and cache"""
    def __init__(self, config_dir, **kwargs):
        # private yaycl conf, plugins can stash info here for configuration
        self._yaycl = AttrDict({'config_dir': config_dir})
        self._yaycl.update(kwargs)
        self._runtime = ConfigTree(self)

    def _runtime_overrides(self):
        return self._runtime

    def _set_runtime_overrides(self, overrides_dict):
        self._runtime.update(overrides_dict)

    def _del_runtime_overrides(self):
        self._runtime.clear()

    runtime = property(_runtime_overrides, _set_runtime_overrides, _del_runtime_overrides)

    def save(self, key):
        """Write out an in-memory config to the conf dir

        Warning: This will destroy any formatting or ordering that existed in the original yaml

        """
        with open(os.path.join(self._yaycl.config_dir, '%s.yaml' % key), 'w') as conf_file:
            self[key].dump(conf_file)

    # Support for descriptor access, e.g. instance.attrname
    # Note that this is only on the get side, for support of nefarious things
    # like setting and deleting, use the normal dict interface.
    def __getattribute__(self, attr):
        # Attempt normal object attr lookup; delegate to the dict interface if that fails
        try:
            return super(Config, self).__getattribute__(attr)
        except AttributeError:
            # try to load from cache first if this conf is already known
            if attr in self:
                return self[attr]

            if attr.startswith('_'):
                # don't try to load private names
                raise

            # If we're here, trigger the loader via getitem
            return self[attr]

    def __getitem__(self, key):
        # Attempt a normal dict lookup to pull a cached conf
        if key not in self:
            super(Config, self).__setitem__(key, AttrDict())
            # Cache miss, populate this conf key
            # Call out to dict's setitem here since this is the only place where we're allowed to
            # create a new key and we want the default behavior instead of the override below
            self._populate(key)
        return super(Config, self).__getitem__(key)

    def __setitem__(self, key, value):
        self[key].clear()
        self[key].update(value)

    def __delitem__(self, key):
        self[key].clear()
        self._populate(key)

    def _inherit(self, conf_key, root=None):
        """Recurses through an object looking for 'inherit' clauses and replaces them with their
        real counterparts. In the case of a dict, the inherit clause remains, in the case of
        anything else, a replacement occurs such that:

        sim5:
          key: value
          newkey: newvalue

        sim6:
          tags:
            - tag1
            - tag2
        sim7:
          inherit: management_systems/sim5
          test:
              tags:
                  inherit: management_systems/sim6/tags

        Will produce the following output if requesting management_systems/sim7

          inherit: management_systems/sim5
          key: value
          newkey: newvalue
          test:
              tags:
                  - tag1
                  - tag2
        """
        if root is None:
            # start at the named conf root if no root is set
            root = self[conf_key]

        for key, value in root.items():
            if key == 'inherit':
                del(root['inherit'])
                self._tree_set(conf_key, root, value)
            try:
                self._inherit(conf_key, value)
            except AttributeError:
                pass

    def _tree_set(self, conf_key, root, path):
        base = self[conf_key]
        for path_element in path.split('/'):
            base = base[path_element]
        root.rebase(base)
        self._inherit(conf_key, root)

    def _populate(self, key):
        yaml_dict = self._load_yaml(key)

        # Graft in local yaml updates if they're available
        with catch_warnings():
            local_yaml = '%s.local' % key
            local_yaml_dict = load_yaml(self, local_yaml, warn_on_fail=False)
            if local_yaml_dict:
                yaml_dict.update_dict(local_yaml_dict)

        # Graft on the runtime overrides
        yaml_dict.update(self.runtime.get(key, {}))
        self[key].update(yaml_dict)
        self._inherit(key)

    def clear(self):
        # because of the 'from conf import foo' mechanism, we need to clear each key in-place,
        # and reload the runtime overrides. Once a key is created in this dict, its value MUST NOT
        # change to a different dict object.
        for key in self:
            # clear the conf dict in-place
            self[key].clear()
            self._populate(key)

    def _load_yaml(self, conf_key, warn_on_fail=True):
        # sort the default loader last
        # return the first conf loader that succeeds,
        # otherwise return the last conf that was loaded
        for ep in sorted(iter_entry_points('yaycl.load_yaml'),
                key=lambda ep: ep.name == 'default'):
            loader = ep.load()
            conf = loader(self, conf_key, warn)
            if conf:
                return conf
        else:
            # Since we put default last, the last conf loaded will
            # an empty AttrDict; warn and return
            if warn_on_fail:
                msg = 'Unable to load configuration "{}"'.format(conf_key)
                warn(msg, ConfigNotFound)
            return conf

class ConfigTree(defaultdict):
    """A tree class that knows to clear the config when mutated

    This ensures runtime overrides persist though conf changes while
    still making it easy to nest items/attrs as much as needed to
    override a specific conf branch easily.

    This supports item getters and setters similar to ``AttrDict``,
    but attr getters and setters are currently unsupported.

    """
    def __init__(self, conf, *args, **kwargs):
        self._conf = conf
        tree_constructor = partial(type(self), conf)
        super(ConfigTree, self).__init__(tree_constructor, *args, **kwargs)

    @property
    def _sup(self):
        return super(ConfigTree, self)

    def __setitem__(self, key, value):
        self._sup.__setitem__(key, value)
        self._clear_conf()

    def __delitem__(self, key):
        self._sup.__delitem__(key)
        self._clear_conf()

    def update(self, other):
        self._sup.update(other)
        self._clear_conf()

    def clear(self):
        self._sup.clear()
        self._clear_conf()

    def _clear_conf(self):
        if self._conf:
            self._conf.clear()


def load_yaml(conf, conf_key, warn_on_fail=True):
    loaded_yaml = AttrDict()

    filename = os.path.join(conf._yaycl.config_dir, '{}.yaml'.format(conf_key))

    # Find the requested yaml in the yaml dir
    if os.path.exists(filename):
        with open(filename) as config_fh:
            try:
                loaded_yaml.update(yaml.load(config_fh, Loader=OrderedDictYAMLLoader))
            except:
                warn('Unable to parse configuration file at {}'.format(filename), ConfigInvalid)

    return loaded_yaml

